import os
import platform

import psutil
from opentelemetry import trace
from opentelemetry.exporter.jaeger.thrift import JaegerExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import \
    OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from fedn.common.log_config import logger


class NoopTracer:
    def start_as_current_span(self, name, context=None, kind=None):
        return NoopSpan()


class NoopSpan:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def set_attribute(self, key, value):
        pass

    def add_event(self, name, attributes=None):
        pass

    def end(self, end_time=None):
        pass

    def record_exception(self, exception, attributes=None):
        pass

    def is_recording(self):
        return False

    def __call__(self, func):
        # Makes NoopSpan callable, and it returns the function itself, hence preserving its original functionality
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper


def initialize_tracer():
    if os.getenv("FEDN_TELEMETRY", 'true').lower() in ('false', '0'):
        return NoopTracer()
    else:
        logger.info("Telemetry enabled. Disable by setting FEDN_TELEMETRY=false")

    telemetry_server = os.getenv("FEDN_TELEMETRY_SERVER", 'https://telemetry.fedn.scaleoutsystems.com')
    telemetry_port = os.getenv("FEDN_TELEMETRY_PORT", 443)
    telemetry_service_name = os.getenv("FEDN_TELEMETRY_SERVICE_NAME", "FEDn")

    # Configure the tracer to report data to Jaeger
    trace.set_tracer_provider(
        TracerProvider(
            resource=Resource.create({SERVICE_NAME: telemetry_service_name})
        )
    )

    otlp_exporter = OTLPSpanExporter(
        endpoint=f"{telemetry_server}:{telemetry_port}",  # Default OTLP port for Jaeger, adjust as needed
        # insecure=True  # Use this only if you are not setting up TLS/SSL
    )

    # Attach the exporter to the tracer provider
    trace.get_tracer_provider().add_span_processor(
        BatchSpanProcessor(otlp_exporter)
    )

    return trace.get_tracer(__name__)


def get_context():
    context = {
                "fedn": {
                    "version": "0.9.1",
                },
                "hw": {
                    "cpu_count": os.cpu_count(),
                    "total_memory": psutil.virtual_memory().total,
                    "available_memory": psutil.virtual_memory().available,
                    "total_disk_space": psutil.disk_usage('/').total,
                    "available_disk_space": psutil.disk_usage('/').free,
                },
                "platform": {
                    "system": platform.system(),
                    "release": platform.release(),
                    "platform": platform.platform(),
                    "python_implementation": platform.python_implementation(),
                    "python_version": platform.python_version(),
                    "machine": platform.machine(),
                    "architecture": platform.architecture(),
                    "version": platform.uname().version,
                },
            }
    return context


# Initialize tracer
tracer = initialize_tracer()
try:
    with tracer.start_as_current_span("initialize_tracer") as span:
        context = get_context()
        span.set_attribute("context", str(context))
except Exception as e:
    logger.error("Failed to initialize tracer: {}".format(e))


def trace_all_methods(cls):
    def traced_method(method):
        """Wrap the method so that it executes within a traced span."""
        def wrapper(*args, **kwargs):
            with tracer.start_as_current_span(method.__name__) as span:
                # Set the class name attribute on the span
                span.set_attribute("class_name", cls.__name__)
                return method(*args, **kwargs)
        return wrapper

    # Apply the traced_method decorator to each callable method of the class
    for key, method in cls.__dict__.items():
        if callable(method) and not key.startswith('__'):
            # Set the decorated method back on the class
            setattr(cls, key, traced_method(method))
    return cls



from typing import Dict, List

import pymongo
from bson import ObjectId
from pymongo.database import Database

from fedn.network.storage.statestore.stores.store import Store

from .shared import EntityNotFound, from_document


class Combiner:
    def __init__(
            self,
            id: str,
            name: str,
            address: str,
            certificate: str,
            config: dict,
            fqdn: str,
            ip: str,
            key: str,
            parent: dict,
            port: int,
            status: str,
            updated_at: str
    ):
        self.id = id
        self.name = name
        self.address = address
        self.certificate = certificate
        self.config = config
        self.fqdn = fqdn
        self.ip = ip
        self.key = key
        self.parent = parent
        self.port = port
        self.status = status
        self.updated_at = updated_at

    def from_dict(data: dict) -> 'Combiner':
        return Combiner(
            id=str(data['_id']),
            name=data['name'] if 'name' in data else None,
            address=data['address'] if 'address' in data else None,
            certificate=data['certificate'] if 'certificate' in data else None,
            config=data['config'] if 'config' in data else None,
            fqdn=data['fqdn'] if 'fqdn' in data else None,
            ip=data['ip'] if 'ip' in data else None,
            key=data['key'] if 'key' in data else None,
            parent=data['parent'] if 'parent' in data else None,
            port=data['port'] if 'port' in data else None,
            status=data['status'] if 'status' in data else None,
            updated_at=data['updated_at'] if 'updated_at' in data else None
        )


class CombinerStore(Store[Combiner]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def get(self, id: str, use_typing: bool = False) -> Combiner:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the name (property)
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entity
        """
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            document = self.database[self.collection].find_one({'_id': id_obj})
        else:
            document = self.database[self.collection].find_one({'name': id})

        if document is None:
            raise EntityNotFound(f"Entity with (id | name) {id} not found")

        return Combiner.from_dict(document) if use_typing else from_document(document)

    def update(self, id: str, item: Combiner) -> bool:
        raise NotImplementedError("Update not implemented for CombinerStore")

    def add(self, item: Combiner) -> bool:
        raise NotImplementedError("Add not implemented for CombinerStore")

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for CombinerStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, use_typing: bool = False, **kwargs) -> Dict[int, List[Combiner]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param use_typing: Whether to return the entities as typed objects or as dicts
            type: bool
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        response = super().list(limit, skip, sort_key or "updated_at", sort_order, use_typing=use_typing, **kwargs)

        result = [Combiner.from_dict(item) for item in response['result']] if use_typing else response['result']

        return {
            "count": response["count"],
            "result": result
        }

    def count(self, **kwargs) -> int:
        return super().count(**kwargs)



import os

import requests

from fedn.common.telemetry import trace_all_methods

__all__ = ['APIClient']


@trace_all_methods
class APIClient:
    """ An API client for interacting with the statestore and controller.

    :param host: The host of the api server.
    :type host: str
    :param port: The port of the api server.
    :type port: int
    :param secure: Whether to use https.
    :type secure: bool
    :param verify: Whether to verify the server certificate.
    :type verify: bool
    """

    def __init__(self, host, port=None, secure=False, verify=False, token=None, auth_scheme=None):
        self.host = host
        self.port = port
        self.secure = secure
        self.verify = verify
        self.headers = {}
        # Auth scheme passed as argument overrides environment variable.
        # "Token" is the default auth scheme.
        if not auth_scheme:
            auth_scheme = os.environ.get("FEDN_AUTH_SCHEME", "Bearer")
        # Override potential env variable if token is passed as argument.
        if not token:
            token = os.environ.get("FEDN_AUTH_TOKEN", False)

        if token:
            self.headers = {"Authorization": f"{auth_scheme} {token}"}

    def _get_url(self, endpoint):
        if self.secure:
            protocol = 'https'
        else:
            protocol = 'http'
        if self.port:
            return f'{protocol}://{self.host}:{self.port}/{endpoint}'
        return f'{protocol}://{self.host}/{endpoint}'

    def _get_url_api_v1(self, endpoint):
        return self._get_url(f'api/v1/{endpoint}')

    # --- Clients --- #

    def get_client(self, id: str):
        """ Get a client from the statestore.

        :param id: The client id to get.
        :type id: str
        :return: Client.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f'clients/{id}'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_clients(self, n_max: int = None):
        """ Get clients from the statestore.

        :param n_max: The maximum number of clients to get (If none all will be fetched).
        :type n_max: int
        return: Clients.
        rtype: dict
        """
        _headers = self.headers.copy()

        if n_max:
            _headers['X-Limit'] = str(n_max)

        response = requests.get(self._get_url_api_v1('clients'), verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_clients_count(self):
        """ Get the number of clients in the statestore.

        :return: The number of clients.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1('clients/count'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_client_config(self, checksum=True):
        """ Get client config from controller. Optionally include the checksum.
        The config is used for clients to connect to the controller and ask for combiner assignment.

        :param checksum: Whether to include the checksum of the package.
        :type checksum: bool
        :return: The client configuration.
        :rtype: dict
        """
        _params = {
            'checksum': "true" if checksum else "false"
        }

        response = requests.get(self._get_url('get_client_config'), params=_params, verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_active_clients(self, combiner_id: str = None, n_max: int = None):
        """ Get active clients from the statestore.

        :param combiner_id: The combiner id to get active clients for.
        :type combiner_id: str
        :param n_max: The maximum number of clients to get (If none all will be fetched).
        :type n_max: int
        :return: Active clients.
        :rtype: dict
        """
        _params = {"status": "online"}

        if combiner_id:
            _params['combiner'] = combiner_id

        _headers = self.headers.copy()

        if n_max:
            _headers['X-Limit'] = str(n_max)

        response = requests.get(self._get_url_api_v1('clients/'), params=_params, verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    # --- Combiners --- #

    def get_combiner(self, id: str):
        """ Get a combiner from the statestore.

        :param id: The combiner id to get.
        :type id: str
        :return: Combiner.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f'combiners/{id}'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_combiners(self, n_max: int = None):
        """ Get combiners in the network.

        :param n_max: The maximum number of combiners to get (If none all will be fetched).
        :type n_max: int
        :return: Combiners.
        :rtype: dict
        """
        _headers = self.headers.copy()

        if n_max:
            _headers['X-Limit'] = str(n_max)

        response = requests.get(self._get_url_api_v1('combiners/'), verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_combiners_count(self):
        """ Get the number of combiners in the statestore.

        :return: The number of combiners.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1('combiners/count'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    # --- Controllers --- #

    def get_controller_status(self):
        """ Get the status of the controller.

        :return: The status of the controller.
        :rtype: dict
        """
        response = requests.get(self._get_url('get_controller_status'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    # --- Models --- #

    def get_model(self, id: str):
        """ Get a model from the statestore.

        :param id: The id (or model property) of the model to get.
        :type id: str
        :return: Model.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f'models/{id}'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_models(self, session_id: str = None, n_max: int = None):
        """ Get models from the statestore.

        :param session_id: The session id to get models for. (optional)
        :type session_id: str
        :param n_max: The maximum number of models to get (If none all will be fetched).
        :type n_max: int
        :return: Models.
        :rtype: dict
        """
        _params = {}

        if session_id:
            _params['session_id'] = session_id

        _headers = self.headers.copy()

        if n_max:
            _headers['X-Limit'] = str(n_max)

        response = requests.get(self._get_url_api_v1('models/'), params=_params, verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_models_count(self):
        """ Get the number of models in the statestore.

        :return: The number of models.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1('models/count'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_active_model(self):
        """ Get the latest model from the statestore.

        :return: The latest model.
        :rtype: dict
        """
        _headers = self.headers.copy()
        _headers['X-Limit'] = "1"

        response = requests.get(self._get_url_api_v1('models/'), verify=self.verify, headers=_headers)
        _json = response.json()

        if "result" in _json and len(_json["result"]) > 0:
            return _json["result"][0]

        return _json

    def get_model_trail(self, id: str = None, include_self: bool = True, reverse: bool = True, n_max: int = None):
        """ Get the model trail.

        :param id: The id (or model property) of the model to start the trail from. (optional)
        :type id: str
        :param n_max: The maximum number of models to get (If none all will be fetched).
        :type n_max: int
        :return: Models.
        :rtype: dict
        """
        if not id:
            model = self.get_active_model()
            if "id" in model:
                id = model["id"]
            else:
                return model

        _headers = self.headers.copy()

        _count: int = n_max if n_max else self.get_models_count()
        _headers['X-Limit'] = str(_count)
        _headers['X-Reverse'] = "true" if reverse else "false"

        _include_self_str: str = "true" if include_self else "false"

        response = requests.get(self._get_url_api_v1(f'models/{id}/ancestors?include_self={_include_self_str}'), verify=self.verify, headers=_headers)
        _json = response.json()

        return _json

    def download_model(self, id: str, path: str):
        """ Download the model with id id.

        :param id: The id (or model property) of the model to download.
        :type id: str
        :param path: The path to download the model to.
        :type path: str
        :return: Message with success or failure.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f'models/{id}/download'), verify=self.verify, headers=self.headers)

        if response.status_code == 200:

            with open(path, 'wb') as file:
                file.write(response.content)
            return {'success': True, 'message': 'Model downloaded successfully.'}
        else:
            return {'success': False, 'message': 'Failed to download model.'}

    def set_active_model(self, path):
        """ Set the initial model in the statestore and upload to model repository.

        :param path: The file path of the initial model to set.
        :type path: str
        :return: A dict with success or failure message.
        :rtype: dict
        """
        with open(path, 'rb') as file:
            response = requests.post(self._get_url('set_initial_model'), files={'file': file}, verify=self.verify, headers=self.headers)
        return response.json()

    # --- Packages --- #

    def get_package(self, id: str):
        """ Get a compute package from the statestore.

        :param id: The id of the compute package to get.
        :type id: str
        :return: Package.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f'packages/{id}'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_packages(self, n_max: int = None):
        """ Get compute packages from the statestore.

        :param n_max: The maximum number of packages to get (If none all will be fetched).
        :type n_max: int
        :return: Packages.
        :rtype: dict
        """
        _headers = self.headers.copy()

        if n_max:
            _headers['X-Limit'] = str(n_max)

        response = requests.get(self._get_url_api_v1('packages/'), verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_packages_count(self):
        """ Get the number of compute packages in the statestore.

        :return: The number of packages.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1('packages/count'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_active_package(self):
        """ Get the (active) compute package from the statestore.

        :return: Package.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1('packages/active'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_package_checksum(self):
        """ Get the checksum of the compute package.

        :return: The checksum.
        :rtype: dict
        """
        response = requests.get(self._get_url('get_package_checksum'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def download_package(self, path: str):
        """ Download the compute package.

        :param path: The path to download the compute package to.
        :type path: str
        :return: Message with success or failure.
        :rtype: dict
        """
        response = requests.get(self._get_url('download_package'), verify=self.verify, headers=self.headers)
        if response.status_code == 200:
            with open(path, 'wb') as file:
                file.write(response.content)
            return {'success': True, 'message': 'Package downloaded successfully.'}
        else:
            return {'success': False, 'message': 'Failed to download package.'}

    def set_active_package(self, path: str, helper: str, name: str = None, description: str = None):
        """ Set the compute package in the statestore.

        :param path: The file path of the compute package to set.
        :type path: str
        :param helper: The helper type to use.
        :type helper: str
        :return: A dict with success or failure message.
        :rtype: dict
        """
        with open(path, 'rb') as file:
            response = requests.post(self._get_url('set_package'), files={'file': file}, data={
                                     'helper': helper, 'name': name, 'description': description}, verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    # --- Rounds --- #

    def get_round(self, id: str):
        """ Get a round from the statestore.

        :param round_id: The round id to get.
        :type round_id: str
        :return: Round (config and metrics).
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f'rounds/{id}'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_rounds(self, n_max: int = None):
        """ Get all rounds from the statestore.

        :param n_max: The maximum number of rounds to get (If none all will be fetched).
        :type n_max: int
        :return: Rounds.
        :rtype: dict
        """
        _headers = self.headers.copy()

        if n_max:
            _headers['X-Limit'] = str(n_max)

        response = requests.get(self._get_url_api_v1('rounds/'), verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_rounds_count(self):
        """ Get the number of rounds in the statestore.

        :return: The number of rounds.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1('rounds/count'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    # --- Sessions --- #

    def get_session(self, id: str):
        """ Get a session from the statestore.

        :param id: The session id to get.
        :type id: str
        :return: Session.
        :rtype: dict
        """

        response = requests.get(self._get_url_api_v1(f'sessions/{id}'), self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_sessions(self, n_max: int = None):
        """ Get sessions from the statestore.

        :param n_max: The maximum number of sessions to get (If none all will be fetched).
        :type n_max: int
        :return: Sessions.
        :rtype: dict
        """
        _headers = self.headers.copy()

        if n_max:
            _headers['X-Limit'] = str(n_max)

        response = requests.get(self._get_url_api_v1('sessions/'), verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_sessions_count(self):
        """ Get the number of sessions in the statestore.

        :return: The number of sessions.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1('sessions/count'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_session_status(self, id: str):
        """ Get the status of a session.

        :param id: The id of the session to get.
        :type id: str
        :return: The status of the session.
        :rtype: str
        """
        session = self.get_session(id)

        if session and "status" in session:
            return session["status"]

        return "Could not retrieve session status."

    def session_is_finished(self, id: str):
        """ Check if a session with id has finished.

        :param id: The id of the session to get.
        :type id: str
        :return: True if session is finished, otherwise false.
        :rtype: bool
        """
        status = self.get_session_status(id)
        return status and status.lower() == "finished"

    def start_session(
            self,
            id: str = None,
            aggregator: str = 'fedavg',
            model_id:  str = None,
            round_timeout: int = 180,
            rounds: int = 5,
            round_buffer_size: int = -1,
            delete_models: bool = True,
            validate: bool = True, helper: str = 'numpyhelper', min_clients: int = 1, requested_clients: int = 8
    ):
        """ Start a new session.

        :param id: The session id to start.
        :type id: str
        :param aggregator: The aggregator plugin to use.
        :type aggregator: str
        :param model_id: The id of the initial model.
        :type model_id: str
        :param round_timeout: The round timeout to use in seconds.
        :type round_timeout: int
        :param rounds: The number of rounds to perform.
        :type rounds: int
        :param round_buffer_size: The round buffer size to use.
        :type round_buffer_size: int
        :param delete_models: Whether to delete models after each round at combiner (save storage).
        :type delete_models: bool
        :param validate: Whether to validate the model after each round.
        :type validate: bool
        :param helper: The helper type to use.
        :type helper: str
        :param min_clients: The minimum number of clients required.
        :type min_clients: int
        :param requested_clients: The requested number of clients.
        :type requested_clients: int
        :return: A dict with success or failure message and session config.
        :rtype: dict
        """
        response = requests.post(self._get_url('start_session'), json={
            'session_id': id,
            'aggregator': aggregator,
            'model_id': model_id,
            'round_timeout': round_timeout,
            'rounds': rounds,
            'round_buffer_size': round_buffer_size,
            'delete_models': delete_models,
            'validate': validate,
            'helper': helper,
            'min_clients': min_clients,
            'requested_clients': requested_clients
        }, verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    # --- Statuses --- #

    def get_status(self, id: str):
        """ Get a status object (event) from the statestore.

        :param id: The id of the status to get.
        :type id: str
        :return: Status.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f'statuses/{id}'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_statuses(self, session_id: str = None, event_type: str = None, sender_name: str = None, sender_role: str = None, n_max: int = None):
        """ Get statuses from the statestore. Filter by input parameters

        :param session_id: The session id to get statuses for.
        :type session_id: str
        :param event_type: The event type to get.
        :type event_type: str
        :param sender_name: The sender name to get.
        :type sender_name: str
        :param sender_role: The sender role to get.
        :type sender_role: str
        :param n_max: The maximum number of statuses to get (If none all will be fetched).
        :type n_max: int
        :return: Statuses
        """
        _params = {}

        if session_id:
            _params["session_id"] = session_id

        if event_type:
            _params["type"] = event_type

        if sender_name:
            _params["sender.name"] = sender_name

        if sender_role:
            _params["sender.role"] = sender_role

        _headers = self.headers.copy()

        if n_max:
            _headers['X-Limit'] = str(n_max)

        response = requests.get(self._get_url_api_v1('statuses/'), params=_params, verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_statuses_count(self):
        """ Get the number of statuses in the statestore.

        :return: The number of statuses.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1('statuses/count'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    # --- Validations --- #

    def get_validation(self, id: str):
        """ Get a validation from the statestore.

        :param id: The id of the validation to get.
        :type id: str
        :return: Validation.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f'validations/{id}'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_validations(
        self,
        session_id: str = None,
        model_id: str = None,
        correlation_id: str = None,
        sender_name: str = None,
        sender_role: str = None,
        receiver_name: str = None,
        receiver_role: str = None,
        n_max: int = None
    ):
        """ Get validations from the statestore. Filter by input parameters.

        :param session_id: The session id to get validations for.
        :type session_id: str
        :param model_id: The model id to get validations for.
        :type model_id: str
        :param correlation_id: The correlation id to get validations for.
        :type correlation_id: str
        :param sender_name: The sender name to get validations for.
        :type sender_name: str
        :param sender_role: The sender role to get validations for.
        :type sender_role: str
        :param receiver_name: The receiver name to get validations for.
        :type receiver_name: str
        :param receiver_role: The receiver role to get validations for.
        :type receiver_role: str
        :param n_max: The maximum number of validations to get (If none all will be fetched).
        :type n_max: int
        :return: Validations.
        :rtype: dict
        """
        _params = {}

        if session_id:
            _params["sessionId"] = session_id

        if model_id:
            _params["modelId"] = model_id

        if correlation_id:
            _params["correlationId"] = correlation_id

        if sender_name:
            _params["sender.name"] = sender_name

        if sender_role:
            _params["sender.role"] = sender_role

        if receiver_name:
            _params["receiver.name"] = receiver_name

        if receiver_role:
            _params["receiver.role"] = receiver_role

        _headers = self.headers.copy()

        if n_max:
            _headers['X-Limit'] = str(n_max)

        response = requests.get(self._get_url_api_v1('validations/'), params=_params, verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_validations_count(self):
        """ Get the number of validations in the statestore.

        :return: The number of validations.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1('validations/count'), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json



import base64
import io
import json
import os
import queue
import re
import socket
import sys
import threading
import time
import uuid
from datetime import datetime
from distutils.dir_util import copy_tree
from io import BytesIO

import grpc
from cryptography.hazmat.primitives.serialization import Encoding
from google.protobuf.json_format import MessageToJson
from OpenSSL import SSL
from tenacity import retry, stop_after_attempt

import fedn.network.grpc.fedn_pb2 as fedn
import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.config import FEDN_AUTH_SCHEME, FEDN_PACKAGE_EXTRACT_DIR
from fedn.common.log_config import logger, set_log_level_from_string, set_log_stream
from fedn.network.clients.connect import ConnectorClient, Status
from fedn.network.clients.package import PackageRuntime
from fedn.network.clients.state import ClientState, ClientStateToString
from fedn.network.combiner.modelservice import get_tmp_path, upload_request_generator
from fedn.utils.dispatcher import Dispatcher
from fedn.utils.helpers.helpers import get_helper

CHUNK_SIZE = 1024 * 1024
VALID_NAME_REGEX = "^[a-zA-Z0-9_-]*$"


class GrpcAuth(grpc.AuthMetadataPlugin):
    def __init__(self, key):
        self._key = key

    def __call__(self, context, callback):
        callback((("authorization", f"{FEDN_AUTH_SCHEME} {self._key}"),), None)


class Client:
    """FEDn Client. Service running on client/datanodes in a federation,
    recieving and handling model update and model validation requests.

    :param config: A configuration dictionary containing connection information for the discovery service (controller)
        and settings governing e.g. client-combiner assignment behavior.
    :type config: dict
    """

    def __init__(self, config):
        """Initialize the client."""
        self.state = None
        self.error_state = False
        self._connected = False
        self._missed_heartbeat = 0
        self.config = config
        self.trace_attribs = False
        set_log_level_from_string(config.get("verbosity", "INFO"))
        set_log_stream(config.get("logfile", None))

        self.connector = ConnectorClient(
            host=config["discover_host"],
            port=config["discover_port"],
            token=config["token"],
            name=config["name"],
            remote_package=config["remote_compute_context"],
            force_ssl=config["force_ssl"],
            verify=config["verify"],
            combiner=config["preferred_combiner"],
            id=config["client_id"],
        )

        # Validate client name
        match = re.search(VALID_NAME_REGEX, config["name"])
        if not match:
            raise ValueError("Unallowed character in client name. Allowed characters: a-z, A-Z, 0-9, _, -.")

        # Folder where the client will store downloaded compute package and logs
        self.name = config["name"]
        if FEDN_PACKAGE_EXTRACT_DIR:
            self.run_path = os.path.join(os.getcwd(), FEDN_PACKAGE_EXTRACT_DIR)
        else:
            dirname = self.name + "-" + time.strftime("%Y%m%d-%H%M%S")
            self.run_path = os.path.join(os.getcwd(), dirname)
        if not os.path.exists(self.run_path):
            os.mkdir(self.run_path)

        self.started_at = datetime.now()
        self.logs = []

        self.inbox = queue.Queue()

        # Attach to the FEDn network (get combiner)
        combiner_config = self.assign()
        self.connect(combiner_config)

        self._initialize_dispatcher(self.config)

        self._initialize_helper(combiner_config)
        if not self.helper:
            logger.warning("Failed to retrieve helper class settings: {}".format(combiner_config))

        self._subscribe_to_combiner(self.config)

        self.state = ClientState.idle

    def assign(self):
        """Contacts the controller and asks for combiner assignment.

        :return: A configuration dictionary containing connection information for combiner.
        :rtype: dict
        """

        logger.info("Initiating assignment request.")
        while True:
            status, response = self.connector.assign()
            if status == Status.TryAgain:
                logger.warning(response)
                logger.info("Assignment request failed. Retrying in 5 seconds.")
                time.sleep(5)
                continue
            if status == Status.Assigned:
                combiner_config = response
                break
            if status == Status.UnAuthorized:
                logger.critical(response)
                sys.exit("Exiting: Unauthorized")
            if status == Status.UnMatchedConfig:
                logger.critical(response)
                sys.exit("Exiting: UnMatchedConfig")
            time.sleep(5)
        logger.info("Assignment successfully received.")
        logger.info("Received combiner configuration: {}".format(combiner_config))
        return combiner_config

    def _add_grpc_metadata(self, key, value):
        """Add metadata for gRPC calls.

        :param key: The key of the metadata.
        :type key: str
        :param value: The value of the metadata.
        :type value: str
        """
        # Check if metadata exists and add if not
        if not hasattr(self, "metadata"):
            self.metadata = ()

        # Check if metadata key already exists and replace value if so
        for i, (k, v) in enumerate(self.metadata):
            if k == key:
                # Replace value
                self.metadata = self.metadata[:i] + ((key, value),) + self.metadata[i + 1 :]
                return

        # Set metadata using tuple concatenation
        self.metadata += ((key, value),)

    def _get_ssl_certificate(self, domain, port=443):
        context = SSL.Context(SSL.SSLv23_METHOD)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((domain, port))
        ssl_sock = SSL.Connection(context, sock)
        ssl_sock.set_tlsext_host_name(domain.encode())
        ssl_sock.set_connect_state()
        ssl_sock.do_handshake()
        cert = ssl_sock.get_peer_certificate()
        ssl_sock.close()
        sock.close()
        cert = cert.to_cryptography().public_bytes(Encoding.PEM).decode()
        return cert

    def connect(self, combiner_config):
        """Connect to combiner.

        :param combiner_config: connection information for the combiner.
        :type combiner_config: dict
        """

        if self._connected:
            logger.info("Client is already attached. ")
            return None

        # TODO use the combiner_config['certificate'] for setting up secure comms'
        host = combiner_config["host"]
        # Add host to gRPC metadata
        self._add_grpc_metadata("grpc-server", host)
        logger.debug("Client using metadata: {}.".format(self.metadata))
        port = combiner_config["port"]
        secure = False
        if combiner_config["fqdn"] is not None:
            host = combiner_config["fqdn"]
            # assuming https if fqdn is used
            port = 443
        logger.info(f"Initiating connection to combiner host at: {host}:{port}")

        if combiner_config["certificate"]:
            logger.info("Utilizing CA certificate for GRPC channel authentication.")
            secure = True
            cert = base64.b64decode(combiner_config["certificate"])  # .decode('utf-8')
            credentials = grpc.ssl_channel_credentials(root_certificates=cert)
            channel = grpc.secure_channel("{}:{}".format(host, str(port)), credentials)
        elif os.getenv("FEDN_GRPC_ROOT_CERT_PATH"):
            secure = True
            logger.info("Using root certificate from environment variable for GRPC channel.")
            with open(os.environ["FEDN_GRPC_ROOT_CERT_PATH"], "rb") as f:
                credentials = grpc.ssl_channel_credentials(f.read())
            channel = grpc.secure_channel("{}:{}".format(host, str(port)), credentials)
        elif self.config["secure"]:
            secure = True
            logger.info("Using CA certificate for GRPC channel.")
            cert = self._get_ssl_certificate(host, port=port)

            credentials = grpc.ssl_channel_credentials(cert.encode("utf-8"))
            if self.config["token"]:
                token = self.config["token"]
                auth_creds = grpc.metadata_call_credentials(GrpcAuth(token))
                channel = grpc.secure_channel("{}:{}".format(host, str(port)), grpc.composite_channel_credentials(credentials, auth_creds))
            else:
                channel = grpc.secure_channel("{}:{}".format(host, str(port)), credentials)
        else:
            logger.info("Using insecure GRPC channel.")
            if port == 443:
                port = 80
            channel = grpc.insecure_channel("{}:{}".format(host, str(port)))

        self.channel = channel

        self.connectorStub = rpc.ConnectorStub(channel)
        self.combinerStub = rpc.CombinerStub(channel)
        self.modelStub = rpc.ModelServiceStub(channel)

        logger.info("Successfully established {} connection to {}:{}".format("secure" if secure else "insecure", host, port))

        logger.info("Using {} compute package.".format(combiner_config["package"]))

        self._connected = True

    def disconnect(self):
        """Disconnect from the combiner."""
        if not self._connected:
            logger.info("Client is not connected.")

        self.channel.close()
        self._connected = False
        logger.info("Client {} disconnected.".format(self.name))

    def _initialize_helper(self, combiner_config):
        """Initialize the helper class for the client.

        :param combiner_config: A configuration dictionary containing connection information for
        | the discovery service (controller) and settings governing e.g.
        | client-combiner assignment behavior.
        :type combiner_config: dict
        :return:
        """

        if "helper_type" in combiner_config.keys():
            self.helper = get_helper(combiner_config["helper_type"])

    def _subscribe_to_combiner(self, config):
        """Listen to combiner message stream and start all processing threads.

        :param config: A configuration dictionary containing connection information for
        | the discovery service (controller) and settings governing e.g.
        | client-combiner assignment behavior.
        """

        # Start sending heartbeats to the combiner.
        threading.Thread(target=self._send_heartbeat, kwargs={"update_frequency": config["heartbeat_interval"]}, daemon=True).start()

        # Start listening for combiner training and validation messages
        threading.Thread(target=self._listen_to_task_stream, daemon=True).start()
        self._connected = True

        # Start processing the client message inbox
        threading.Thread(target=self.process_request, daemon=True).start()

    @retry(stop=stop_after_attempt(3))
    def untar_package(self, package_runtime):
        _, package_runpath = package_runtime.unpack()
        return package_runpath

    def _initialize_dispatcher(self, config):
        """Initialize the dispatcher for the client.

        :param config: A configuration dictionary containing connection information for
        | the discovery service (controller) and settings governing e.g.
        | client-combiner assignment behavior.
        :type config: dict
        :return:
        """
        if config["remote_compute_context"]:
            pr = PackageRuntime(self.run_path)

            retval = None
            tries = 10

            while tries > 0:
                retval = pr.download(
                    host=config["discover_host"], port=config["discover_port"], token=config["token"], force_ssl=config["force_ssl"], secure=config["secure"]
                )
                if retval:
                    break
                time.sleep(60)
                logger.warning("Compute package not available. Retrying in 60 seconds. {} attempts remaining.".format(tries))
                tries -= 1

            if retval:
                if "checksum" not in config:
                    logger.warning("Bypassing validation of package checksum. Ensure the package source is trusted.")
                else:
                    checks_out = pr.validate(config["checksum"])
                    if not checks_out:
                        logger.critical("Validation of local package failed. Client terminating.")
                        self.error_state = True
                        return
            package_runpath = ""
            if retval:
                package_runpath = self.untar_package(pr)

            self.dispatcher = pr.dispatcher(package_runpath)
            try:
                logger.info("Initiating Dispatcher with entrypoint set to: startup")
                activate_cmd = self.dispatcher._get_or_create_python_env()
                self.dispatcher.run_cmd("startup")
            except KeyError:
                logger.info("No startup command found in package. Continuing.")
                pass
            except Exception as e:
                logger.error(f"Caught exception: {type(e).__name__}")

        else:
            # TODO: Deprecate
            dispatch_config = {
                "entry_points": {
                    "predict": {"command": "python3 predict.py"},
                    "train": {"command": "python3 train.py"},
                    "validate": {"command": "python3 validate.py"},
                }
            }
            from_path = os.path.join(os.getcwd(), "client")

            copy_tree(from_path, self.run_path)
            self.dispatcher = Dispatcher(dispatch_config, self.run_path)
        # Get or create python environment
        activate_cmd = self.dispatcher._get_or_create_python_env()
        if activate_cmd:
            logger.info("To activate the virtual environment, run: {}".format(activate_cmd))

    def get_model_from_combiner(self, id, timeout=20):
        """Fetch a model from the assigned combiner.
        Downloads the model update object via a gRPC streaming channel.

        :param id: The id of the model update object.
        :type id: str
        :return: The model update object.
        :rtype: BytesIO
        """
        data = BytesIO()
        time_start = time.time()
        request = fedn.ModelRequest(id=id)
        request.sender.name = self.name
        request.sender.role = fedn.WORKER

        try:
            for part in self.modelStub.Download(request, metadata=self.metadata):
                if part.status == fedn.ModelStatus.IN_PROGRESS:
                    data.write(part.data)

                if part.status == fedn.ModelStatus.OK:
                    return data

                if part.status == fedn.ModelStatus.FAILED:
                    return None

                if part.status == fedn.ModelStatus.UNKNOWN:
                    if time.time() - time_start >= timeout:
                        return None
                    continue
        except grpc.RpcError as e:
            logger.critical(f"GRPC: An error occurred during model download: {e}")

        return data

    def send_model_to_combiner(self, model, id):
        """Send a model update to the assigned combiner.
        Uploads the model updated object via a gRPC streaming channel, Upload.

        :param model: The model update object.
        :type model: BytesIO
        :param id: The id of the model update object.
        :type id: str
        :return: The model update object.
        :rtype: BytesIO
        """
        if not isinstance(model, BytesIO):
            bt = BytesIO()

            for d in model.stream(32 * 1024):
                bt.write(d)
        else:
            bt = model

        bt.seek(0, 0)

        try:
            result = self.modelStub.Upload(upload_request_generator(bt, id), metadata=self.metadata)
        except grpc.RpcError as e:
            logger.critical(f"GRPC: An error occurred during model upload: {e}")

        return result

    def _listen_to_task_stream(self):
        """Subscribe to the model update request stream.

        :return: None
        :rtype: None
        """

        r = fedn.ClientAvailableMessage()
        r.sender.name = self.name
        r.sender.role = fedn.WORKER
        # Add client to metadata
        self._add_grpc_metadata("client", self.name)

        while self._connected:
            try:
                for request in self.combinerStub.TaskStream(r, metadata=self.metadata):
                    if request:
                        logger.debug("Received model update request from combiner: {}.".format(request))
                    if request.sender.role == fedn.COMBINER:
                        # Process training request
                        self.send_status(
                            "Received model update request.",
                            log_level=fedn.Status.AUDIT,
                            type=fedn.StatusType.MODEL_UPDATE_REQUEST,
                            request=request,
                            sesssion_id=request.session_id,
                        )
                        logger.info("Received model update request of type {} for model_id {}".format(request.type, request.model_id))

                        if request.type == fedn.StatusType.MODEL_UPDATE and self.config["trainer"]:
                            self.inbox.put(("train", request))
                        elif request.type == fedn.StatusType.MODEL_VALIDATION and self.config["validator"]:
                            self.inbox.put(("validate", request))
                        else:
                            logger.error("Unknown request type: {}".format(request.type))

            except grpc.RpcError as e:
                # Handle gRPC errors
                status_code = e.code()
                if status_code == grpc.StatusCode.UNAVAILABLE:
                    logger.warning("GRPC TaskStream: server unavailable during model update request stream. Retrying.")
                    # Retry after a delay
                    time.sleep(5)
                if status_code == grpc.StatusCode.UNAUTHENTICATED:
                    details = e.details()
                    if details == "Token expired":
                        logger.warning("GRPC TaskStream: Token expired. Reconnecting.")
                        self.detach()

                if status_code == grpc.StatusCode.CANCELLED:
                    # Expected if the client is detached
                    logger.critical("GRPC TaskStream: Client detached from combiner. Atempting to reconnect.")

                else:
                    # Log the error and continue
                    logger.error(f"GRPC TaskStream: An error occurred during model update request stream: {e}")

            except Exception as ex:
                # Handle other exceptions
                logger.error(f"GRPC TaskStream: An error occurred during model update request stream: {ex}")

        # Detach if not attached
        if not self._connected:
            return

    def _process_training_request(self, model_id: str, session_id: str = None):
        """Process a training (model update) request.

        :param model_id: The model id of the model to be updated.
        :type model_id: str
        :param session_id: The id of the current session
        :type session_id: str
        :return: The model id of the updated model, or None if the update failed. And a dict with metadata.
        :rtype: tuple
        """

        self.send_status("\t Starting processing of training request for model_id {}".format(model_id), sesssion_id=session_id)
        self.state = ClientState.training

        try:
            meta = {}
            tic = time.time()
            mdl = self.get_model_from_combiner(str(model_id))
            if mdl is None:
                logger.error("Could not retrieve model from combiner. Aborting training request.")
                return None, None
            meta["fetch_model"] = time.time() - tic

            inpath = self.helper.get_tmp_path()
            with open(inpath, "wb") as fh:
                fh.write(mdl.getbuffer())

            outpath = self.helper.get_tmp_path()
            tic = time.time()
            # TODO: Check return status, fail gracefully

            self.dispatcher.run_cmd("train {} {}".format(inpath, outpath))

            meta["exec_training"] = time.time() - tic

            tic = time.time()
            out_model = None

            with open(outpath, "rb") as fr:
                out_model = io.BytesIO(fr.read())

            # Stream model update to combiner server
            updated_model_id = uuid.uuid4()
            self.send_model_to_combiner(out_model, str(updated_model_id))
            meta["upload_model"] = time.time() - tic

            # Read the metadata file
            with open(outpath + "-metadata", "r") as fh:
                training_metadata = json.loads(fh.read())
            meta["training_metadata"] = training_metadata

            os.unlink(inpath)
            os.unlink(outpath)
            os.unlink(outpath + "-metadata")

        except Exception as e:
            logger.error("Could not process training request due to error: {}".format(e))
            updated_model_id = None
            meta = {"status": "failed", "error": str(e)}

        self.state = ClientState.idle

        return updated_model_id, meta

    def _process_validation_request(self, model_id: str, is_inference: bool, session_id: str = None):
        """Process a validation request.

        :param model_id: The model id of the model to be validated.
        :type model_id: str
        :param is_inference: True if the validation is an inference request, False if it is a validation request.
        :type is_inference: bool
        :param session_id: The id of the current session.
        :type session_id: str
        :return: The validation metrics, or None if validation failed.
        :rtype: dict
        """
        # Figure out cmd
        if is_inference:
            cmd = "infer"
        else:
            cmd = "validate"

        self.send_status(f"Processing {cmd} request for model_id {model_id}", sesssion_id=session_id)
        self.state = ClientState.validating
        try:
            model = self.get_model_from_combiner(str(model_id))
            if model is None:
                logger.error("Could not retrieve model from combiner. Aborting validation request.")
                return None
            inpath = self.helper.get_tmp_path()

            with open(inpath, "wb") as fh:
                fh.write(model.getbuffer())

            outpath = get_tmp_path()
            self.dispatcher.run_cmd(f"{cmd} {inpath} {outpath}")

            with open(outpath, "r") as fh:
                validation = json.loads(fh.read())

            os.unlink(inpath)
            os.unlink(outpath)

        except Exception as e:
            logger.warning("Validation failed with exception {}".format(e))
            self.state = ClientState.idle
            return None

        self.state = ClientState.idle
        return validation

    def process_request(self):
        """Process training and validation tasks."""
        while True:
            if not self._connected:
                return

            try:
                (task_type, request) = self.inbox.get(timeout=1.0)
                if task_type == "train":
                    tic = time.time()
                    self.state = ClientState.training
                    model_id, meta = self._process_training_request(request.model_id, session_id=request.session_id)

                    if meta is not None:
                        processing_time = time.time() - tic
                        meta["processing_time"] = processing_time
                        meta["config"] = request.data

                    if model_id is not None:
                        # Send model update to combiner
                        update = fedn.ModelUpdate()
                        update.sender.name = self.name
                        update.sender.role = fedn.WORKER
                        update.receiver.name = request.sender.name
                        update.receiver.role = request.sender.role
                        update.model_id = request.model_id
                        update.model_update_id = str(model_id)
                        update.timestamp = str(datetime.now())
                        update.correlation_id = request.correlation_id
                        update.meta = json.dumps(meta)

                        try:
                            _ = self.combinerStub.SendModelUpdate(update, metadata=self.metadata)
                            self.send_status(
                                "Model update completed.",
                                log_level=fedn.Status.AUDIT,
                                type=fedn.StatusType.MODEL_UPDATE,
                                request=update,
                                sesssion_id=request.session_id,
                            )
                        except grpc.RpcError as e:
                            status_code = e.code()
                            logger.error("GRPC error, {}.".format(status_code.name))
                            logger.debug(e)
                        except ValueError as e:
                            logger.error("GRPC error, RPC channel closed. {}".format(e))
                            logger.debug(e)
                    else:
                        self.send_status(
                            "Client {} failed to complete model update.", log_level=fedn.Status.WARNING, request=request, sesssion_id=request.session_id
                        )

                    self.state = ClientState.idle
                    self.inbox.task_done()

                elif task_type == "validate":
                    self.state = ClientState.validating
                    metrics = self._process_validation_request(request.model_id, False, request.session_id)

                    if metrics is not None:
                        # Send validation
                        validation = fedn.ModelValidation()
                        validation.sender.name = self.name
                        validation.sender.role = fedn.WORKER
                        validation.receiver.name = request.sender.name
                        validation.receiver.role = request.sender.role
                        validation.model_id = str(request.model_id)
                        validation.data = json.dumps(metrics)
                        validation.timestamp.GetCurrentTime()
                        validation.correlation_id = request.correlation_id
                        validation.session_id = request.session_id

                        try:
                            _ = self.combinerStub.SendModelValidation(validation, metadata=self.metadata)

                            status_type = fedn.StatusType.MODEL_VALIDATION
                            self.send_status(
                                "Model validation completed.", log_level=fedn.Status.AUDIT, type=status_type, request=validation, sesssion_id=request.session_id
                            )
                        except grpc.RpcError as e:
                            status_code = e.code()
                            logger.error("GRPC error, {}.".format(status_code.name))
                            logger.debug(e)
                        except ValueError as e:
                            logger.error("GRPC error, RPC channel closed. {}".format(e))
                            logger.debug(e)
                    else:
                        self.send_status(
                            "Client {} failed to complete model validation.".format(self.name),
                            log_level=fedn.Status.WARNING,
                            request=request,
                            sesssion_id=request.session_id,
                        )

                    self.state = ClientState.idle
                    self.inbox.task_done()
            except queue.Empty:
                pass
            except grpc.RpcError as e:
                logger.critical(f"GRPC process_request: An error occurred during process request: {e}")

    def _send_heartbeat(self, update_frequency=2.0):
        """Send a heartbeat to the combiner.

        :param update_frequency: The frequency of the heartbeat in seconds.
        :type update_frequency: float
        :return: None if the client is detached.
        :rtype: None
        """
        while True:
            heartbeat = fedn.Heartbeat(sender=fedn.Client(name=self.name, role=fedn.WORKER))
            try:
                self.connectorStub.SendHeartbeat(heartbeat, metadata=self.metadata)
                self._missed_heartbeat = 0
            except grpc.RpcError as e:
                status_code = e.code()
                if status_code == grpc.StatusCode.UNAVAILABLE:
                    self._missed_heartbeat += 1
                    logger.error(
                        "GRPC hearbeat: combiner unavailable, retrying (attempt {}/{}).".format(
                            self._missed_heartbeat, self.config["reconnect_after_missed_heartbeat"]
                        )
                    )
                    if self._missed_heartbeat > self.config["reconnect_after_missed_heartbeat"]:
                        self.disconnect()
                if status_code == grpc.StatusCode.UNAUTHENTICATED:
                    details = e.details()
                    if details == "Token expired":
                        logger.error("GRPC hearbeat: Token expired. Disconnecting.")
                        self.disconnect()
                        sys.exit("Unauthorized. Token expired. Please obtain a new token.")
                logger.debug(e)

            time.sleep(update_frequency)
            if not self._connected:
                logger.info("SendStatus: Client disconnected.")
                return

    def send_status(self, msg, log_level=fedn.Status.INFO, type=None, request=None, sesssion_id: str = None):
        """Send status message.

        :param msg: The message to send.
        :type msg: str
        :param log_level: The log level of the message.
        :type log_level: fedn.Status.INFO, fedn.Status.WARNING, fedn.Status.ERROR
        :param type: The type of the message.
        :type type: str
        :param request: The request message.
        :type request: fedn.Request
        """

        if not self._connected:
            logger.info("SendStatus: Client disconnected.")
            return

        status = fedn.Status()
        status.timestamp.GetCurrentTime()
        status.sender.name = self.name
        status.sender.role = fedn.WORKER
        status.log_level = log_level
        status.status = str(msg)
        status.session_id = sesssion_id
        if type is not None:
            status.type = type

        if request is not None:
            status.data = MessageToJson(request)

        self.logs.append("{} {} LOG LEVEL {} MESSAGE {}".format(str(datetime.now()), status.sender.name, status.log_level, status.status))
        try:
            _ = self.connectorStub.SendStatus(status, metadata=self.metadata)
        except grpc.RpcError as e:
            status_code = e.code()
            if status_code == grpc.StatusCode.UNAVAILABLE:
                logger.warning("GRPC SendStatus: server unavailable during send status.")
            if status_code == grpc.StatusCode.UNAUTHENTICATED:
                details = e.details()
                if details == "Token expired":
                    logger.warning("GRPC SendStatus: Token expired.")

    def run(self):
        """Run the client."""
        try:
            cnt = 0
            old_state = self.state
            while True:
                time.sleep(1)
                if cnt == 0:
                    logger.info("Client is active, waiting for model update requests.")
                    cnt = 1
                if self.state != old_state:
                    logger.info("Client in {} state.".format(ClientStateToString(self.state)))
                if not self._connected:
                    logger.warning("Client lost connection to combiner. Attempting to reconnect to FEDn network.")
                    combiner_config = self.assign()
                    self.connect(combiner_config)
                    self._subscribe_to_combiner(self.config)
                    cnt = 0
                if self.error_state:
                    logger.error("Client in error state. Terminiating.")
                    sys.exit("Client in error state. Terminiating.")
        except KeyboardInterrupt:
            logger.info("Shutting down.")
